/* 
 * Copyright (c) 2003 by c0ffee, distributed under the GPL
 *
 * cc rproxy.c -o rproxy
 *
 * may require -lresolv
 *
 * may be redistributed freely as long as the above copyright notice is
 * left intact.
 */
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/types.h>
#include <string.h>
#include <signal.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <errno.h>

static int forward_port;

#undef max
#define max(x,y) ((x) > (y) ? (x) : (y))

static int connect_socket (int connect_port, char *address) 
{
    struct sockaddr_in a;
    int s;
    if ((s = socket (AF_INET, SOCK_STREAM, 0)) < 0) {
        perror ("socket");
        close (s);
        return -1;
    }

    memset (&a, 0, sizeof (a));
    a.sin_port = htons (connect_port);
    a.sin_family = AF_INET;

    if (!inet_aton (address, (struct in_addr *) &a.sin_addr.s_addr)) {
        perror ("bad IP address format");
        close (s);
        return -1;
    }

    if (connect(s, (struct sockaddr *) &a, sizeof (a)) < 0) {
        perror ("connect()");
        shutdown (s, SHUT_RDWR);
        close (s);
        return -1;
    }
    return s;
}

#define BUF_SIZE 8092

int main (int argc, char **argv) {
    int h;
    int fd1 = -1, fd2 = -1;
    char buf1[BUF_SIZE], buf2[BUF_SIZE];
    int buf1_avail, buf1_written;
    int buf2_avail, buf2_written;

    if (argc != 5) {
        fprintf (stderr, "Usage\n\tsquid-ssh <proxy-ip> <proxy-port> <host-name> <host-port>\n");
        exit (1);
    }

    signal (SIGPIPE, SIG_IGN);

    forward_port = atoi (argv[2]);

    fd1 = STDIN_FILENO;
    buf1_avail = buf1_written = 0;
    buf2_avail = buf2_written = 0;
    fd2 = connect_socket (forward_port, argv[1]);
    if (fd2 < 0)
    	exit (1);

    {
        int r;
        char *msg;

        #define SQUID_CMD "CONNECT %s:%s HTTP/1.1\n" \
        "User-Agent: Mozilla/5.0 (X11; U; Linux i686; en-US; rv:1.2.1) Gecko/20030120\n" \
        "Host: %s\n" \
        "\n"

        msg = calloc(strlen(SQUID_CMD) + strlen(argv[3])*2 + strlen(argv[4]) + 1, sizeof(char));
        sprintf (msg, SQUID_CMD, argv[3], argv[4], argv[3]);
        if (write (fd2, msg, strlen (msg)) != strlen (msg))  {
            perror("write()");
	    exit (1);
        }

	free (msg);

    }

    for (;;) {
        int r, nfds = 0;
        fd_set rd, wr, er;
        FD_ZERO (&rd);
        FD_ZERO (&wr);
        FD_ZERO (&er);
        FD_SET (fd1, &er);
        nfds = max (nfds, fd1);
        FD_SET (fd2, &er);
        nfds = max (nfds, fd2);
        if (buf1_avail < BUF_SIZE) {
            FD_SET (fd1, &rd);
            nfds = max (nfds, fd1);
        }
        if (buf2_avail < BUF_SIZE) {
            FD_SET (fd2, &rd);
            nfds = max (nfds, fd2);
        }
        if (buf2_avail - buf2_written > 0) {
            FD_SET (STDOUT_FILENO, &wr);
            nfds = max (nfds, STDOUT_FILENO);
        }
        if (buf1_avail - buf1_written > 0) {
            FD_SET (fd2, &wr);
            nfds = max (nfds, fd2);
        }

        r = select (nfds + 1, &rd, &wr, &er, NULL);

        if (r == -1 && errno == EINTR)
            continue;
        if (r < 0) {
            perror ("select()");
            exit (1);
        }
/* NB: read oob data before normal reads */
        if (FD_ISSET (fd1, &er)) {
            char c;
            errno = 0;
            r = recv (fd1, &c, 1, MSG_OOB);
            if (r < 1) {
	        if (r < 0)
		    perror("local error");
                exit (1);
            } else
                send (fd2, &c, 1, MSG_OOB);
        }
        if (FD_ISSET (fd2, &er)) {
            char c;
            errno = 0;
            r = recv (fd2, &c, 1, MSG_OOB);
            if (r < 1) {
	        if (r < 0)
	            perror("remote error");
	        exit (1);
            } else
                send (fd1, &c, 1, MSG_OOB);
        }
        if (FD_ISSET (fd1, &rd)) {
            r = read (fd1, buf1 + buf1_avail, BUF_SIZE - buf1_avail);
            if (r < 1) {
	        if (r < 0)
                    perror("local read()");
		exit (1);
            } else
                buf1_avail += r;
        }
        if (FD_ISSET (fd2, &rd)) {
            r = read (fd2, buf2 + buf2_avail, BUF_SIZE - buf2_avail);
            if (r < 1) {
	        if (r < 0)
                    perror("remote read()");
		exit (1);
            } else
                buf2_avail += r;
        }
        if (FD_ISSET (STDOUT_FILENO, &wr)) {
            r = write (STDOUT_FILENO, buf2 + buf2_written, buf2_avail - buf2_written);
            if (r < 1) {
	        if (r < 0)
                    perror("local write()");
		exit (1);
            } else
                buf2_written += r;
        }
        if (FD_ISSET (fd2, &wr)) {
            r = write (fd2, buf1 + buf1_written, buf1_avail - buf1_written);
            if (r < 1) {
	        if (r < 0)
	            perror("remote write()");
		exit (1);
            } else
                buf1_written += r;
        }
/* check if write data has caught read data */
        if (buf1_written == buf1_avail)
            buf1_written = buf1_avail = 0;
        if (buf2_written == buf2_avail)
            buf2_written = buf2_avail = 0;
    }
    return 0;
}

